"""
Read data in xsf file.

Note:
    - https://web.mit.edu/xcrysden_v1.5.60/www/XCRYSDEN/doc/XSF.html#__toc__12
"""

import numpy as np
from qtdraw.util.util import create_grid


# ==================================================
def extract_data_xsf(filename):
    """
    Read xsf file (grid data part only).

    Args:
        filename (str): file name.

    Returns:
        - (dict) -- extracted data.

    Note:
        - n : [nx,ny,nz] division of grid.
        - origin : [rx,ry,rz] origin in fractional coordinate.
        - Ag : [g1,g2,g3] grid vectors in 4x4 matrix.
        - data : data at each grid point.
        - surface : surface data at each grid point.
        - endpoint : include endpoint ?
        - row_major : row-major grid ?
    """
    with open(filename, mode="r", encoding="utf-8") as f:
        s = f.read()
    s = s.replace("\n", " ")

    start = s.find("BEGIN_DATAGRID_3D_UNKNOWN")
    end = s.rfind("END_DATAGRID_3D")

    s = s[start:end]
    lst = s.split(" ")
    lst = [i for i in lst if i != ""]

    nv = np.array(lst[1:4]).astype(int)  # nx, ny, nz
    rv = np.array(lst[4:7]).astype(float)  # x0, y0, z0
    A = np.eye(4)
    A[0:3, 0] = np.array(lst[7:10]).astype(float)  # dx1, dy1, dz1
    A[0:3, 1] = np.array(lst[10:13]).astype(float)  # dx2, dy2, dz2
    A[0:3, 2] = np.array(lst[13:16]).astype(float)  # dx3, dy3, dz3
    n = nv[0] * nv[1] * nv[2]
    data = np.array(lst[16 : 16 + n]).astype(float)  # data (nx*ny*nz)
    r0 = np.linalg.inv(A)[0:3, 0:3] @ rv

    grid_data = {
        "n": nv.tolist(),
        "origin": r0.tolist(),
        "Ag": A.tolist(),
        "endpoint": True,
        "row_major": False,
        "data": np.abs(data).tolist(),
        "surface": {"phase": data.tolist()},
    }

    return grid_data


# ==================================================
def create_data(n, origin, A, endpoint, f_data, f_surface):
    r = origin + np.array([1.0, 1.0, 1.0])
    grid = create_grid(n, origin, r, A, endpoint).points

    data = f_data(grid[:, 0], grid[:, 1], grid[:, 2]).tolist()

    if f_surface is not None:
        surface = {}
        for name, f in f_surface.items():
            surface[name] = f(grid[:, 0], grid[:, 1], grid[:, 2]).tolist()
    else:
        surface = None

    grid_data = {
        "n": n,
        "origin": origin,
        "Ag": A.tolist(),
        "data": data,
        "surface": surface,
        "endpoint": endpoint,
        "row_major": False,
    }

    return grid_data



"""
Tab of GroupView.

This module provides a set of group view classes in tab format.
"""

from PySide6.QtWidgets import QTabWidget, QDialog
from PySide6.QtGui import QFont
from qtdraw.widget.group_view import GroupView
from qtdraw.widget.custom_widget import Layout


# ==================================================
class TabGroupView(QDialog):
    # ==================================================
    def __init__(
        self,
        models,
        title="Data",
        parent=None,
    ):
        """
        Data view group.

        Args:
            models (dict): set of models, {object_type: GroupModel}.
            title (str, optional): window title.
            parent (QWidget, optional): parent.
        """
        super().__init__(parent=parent)
        self.setWindowTitle(title)

        self.tab = QTabWidget(self)
        self.tab.setUsesScrollButtons(True)
        bold = QFont()
        bold.setBold(True)
        self.tab.tabBar().setFont(bold)

        self.view = {}
        for object_type, model in models.items():
            item = GroupView(model, parent=self)
            self.tab.addTab(item, object_type)
            self.view[object_type] = item

        layout = Layout(self)
        layout.addWidget(self.tab, 0, 0, 1, 1)

        self.setLayout(layout)

    # ==================================================
    def select_tab(self, object_type):
        """
        Select tab.

        Args:
            object_type (str): object type.
        """
        idx = list(self.view.keys()).index(object_type)
        self.tab.setCurrentIndex(idx)

    # ==================================================
    def closeEvent(self, event):
        """
        Close event for deselect all.

        :meta private:
        """
        for view in self.view.values():
            view.closeEvent(event)
        super().closeEvent(event)

    # ==================================================
    def update_widget(self):
        """
        Update widget.

        :meta private:
        """
        for view in self.view.values():
            view.update_widget(force=True)



"""
Utility.

This module contains the versatile utilities.
"""

import ast
from pathlib import Path
import pyvista as pv
import numpy as np
import matplotlib.pyplot as plt
from gcoreutils.convert_util import text_to_list, apply
from qtdraw.util.util_str import str_to_sympy1


# ==================================================
def convert_to_str(v):
    """
    Convert from object to str, and remove spaces.

    Args:
        v (Any): object.

    Returns:
        - (str) -- converted str.
    """
    return str(v).replace(" ", "").replace("\t", "").replace("\n", "")


# ==================================================
def read_dict(filename):
    """
    Read dict text file.

    Args:
        filename (str): file name.

    Returns:
        - (dict) -- dictionary from dict text.
    """
    with open(filename, mode="r", encoding="utf-8") as f:
        s = f.read()

    if s[: s.find("{")].count("=") > 0:
        s = s.split("=")[-1].strip(" ")

    c = ast.get_docstring(ast.parse(s))
    if c is not None:
        s = s.replace(c, "").replace('"""', "")
    d = ast.literal_eval(s)

    return d


# ==================================================
def write_dict(filename, dic, header=None, var=None):
    """
    write dict text file.

    Args:
        filename (str): filename.
        dic (dict): dictionary to write.
        header (str, optional): header comment at the top of file.
        var (str, optional): varialbe of dict.
    """
    with open(filename, mode="w", encoding="utf-8") as f:
        if header is not None:
            print('"""' + header + '"""', file=f)
        if var is None:
            print(dic, file=f)
        else:
            print(f"{var} =", dic, file=f)


# ==================================================
def convert_str_vector(vector, cell="[0,0,0]", transform=True, A=None):
    """
    Convert 3-component vector(s) to A.(position+cell).

    Args:
        vector (str): vector, str([float]) or str([[float]]).
        cell (str, optional): cell, str([int]).
        transform (bool, optional): transform by using A ?
        A (numpy.ndarray, optional): A.

    Returns:
        - (numpy.ndarray) -- transformed position.
    """
    cell = np.array(apply(int, text_to_list(cell)))
    vector = str_to_sympy1(vector).astype(float)

    vectorT = vector + cell
    if transform:
        A = A[0:3, 0:3].T
        vectorT = vectorT @ A

    return vectorT


# ==================================================
def check_multipie():
    """
    Check if multipie is installed or not.

    Returns:
        - (bool) -- installed ?
    """
    try:
        import multipie

        return True
    except ImportError:
        return False


# ==================================================
def set_latex_setting():
    """
    Set LaTeX setting.
    """
    package = ["amsmath", "amssymb", "physics"]
    pre = "\n".join([r"\usepackage{" + i + "}" for i in package])
    tex_setting = {"text.usetex": True, "text.latex.preamble": pre}
    plt.rcParams.update(tex_setting)


# ==================================================
def create_grid(grid_n, grid_min, grid_max, A=None, endpoint=False):
    """
    Create grid.

    Args:
        grid_n (list): grid size.
        grid_min (list): grid minimum.
        grid_max (list): grid maximum.
        A (list): [a1,a2,a3].
        endpoint (bool, optional): include end points ?

    Returns:
        UniformGrid: uniform grid.

    Note:
        - grid in column-major order.
    """
    if A is None:
        A = np.eye(4)
    if endpoint:
        s = [(ma - mi) / (n - 1) for mi, ma, n in zip(grid_min, grid_max, grid_n)]
    else:
        s = [(ma - mi) / n for mi, ma, n in zip(grid_min, grid_max, grid_n)]
    grid = pv.ImageData(dimensions=grid_n, origin=grid_min, spacing=s).cast_to_unstructured_grid()
    grid.transform(A)

    return grid


# ==================================================
def split_filename(filename):
    """
    Split file name.

    Args:
        filename (str): filename.

    Returns:
        - (str) -- filename with absolute path.
        - (str) -- filename with relative path.
        - (str) -- base filename.
        - (str) -- extension.
        - (str) -- directory.
    """
    path = Path(filename)
    path_abs = path if path.is_absolute() else (Path.cwd() / path).resolve()
    path_rel = path_abs.relative_to(Path.cwd())
    base = str(path_rel.stem)
    ext = str(path_rel.suffix)
    folder = str(path_abs.parent)

    return str(path_abs), str(path_rel), base, ext, folder


# ==================================================
def cat_filename(base, ext=None):
    """
    Cat filename.

    Args:
        base (str): (base) filename.
        ext (str, optional): extension.

    Returns:
        - (str) -- full file name.
    """
    if ext is not None:
        base = base + ext
    path = str(Path.cwd() / Path(base))

    return path


# ==================================================
def create_style_sheet(size):
    """
    Create style sheet.

    Args:
        size (int): font size (point).

    Returns:
        - (str) -- style sheet string.
    """
    return f"""
    QTreeView::item {{
        padding: 10px 15px 10px 15px;
        border: none;
        outline: none;
        background: none;
    }}

    QTreeView::item:selected {{
        border: none;
        outline: none;
        color: black;
        background: lightyellow;
    }}
    QTreeView::item:focus {{
        border: none;
        outline: none;
        color: black;
        background: lightyellow;
    }}

    QComboBox {{
        font-size: {size}pt;
        min-width: 40px;
    }}

    QLabel {{
        font-size: {size}pt;
    }}

    QPushButton {{
        font-size: {size}pt;
    }}
    """


# ==================================================
def get_data_range(data):
    v1 = data.min()
    v2 = data.max()
    v = max(abs(v1), abs(v2))
    if v1 * v2 < 0.0:
        clim = [-v, v]
    elif v1 < 0.0:
        clim = [-v, 0.0]
    else:
        clim = [0.0, v]
    return clim



"""
Delegate for GroupModel and GroupView.

This module provides delegate for color selector, combo, and editor.
"""

from PySide6.QtWidgets import QStyle, QStyledItemDelegate
from PySide6.QtCore import Qt
from qtdraw.widget.custom_widget import Combo, Editor, ColorSelector


# ==================================================
class Delegate(QStyledItemDelegate):
    # ==================================================
    def __init__(self, parent):
        """
        Base delegate.

        Args:
            parent (QWidget): parent.
        """
        super().__init__(parent)

    # ==================================================
    def setEditorData(self, editor, index):
        """
        Set editor data.

        Args:
            editor (QWidget): editor.
            index (QModelIndex): index.
        """
        value = index.model().data(index, Qt.EditRole)
        editor.setCurrentText(value)

    # ==================================================
    def setModelData(self, editor, model, index):
        """
        Set model data.

        Args:
            editor (QWidget): editor.
            model (QAbstractItemModel) model.
            index (QModelIndex): index.
        """
        model.blockSignals(True)
        model.setData(index, editor.currentText(), Qt.EditRole)
        model.blockSignals(False)

    # ==================================================
    def updateEditorGeometry(self, editor, option, index):
        """
        Update editor geometry.

        Args:
            editor (QWidget): editor.
            option (QStyleOptionViewItem) option.
            index (QModelIndex): index.
        """
        editor.setGeometry(option.rect)

    # ==================================================
    def paint(self, painter, option, index):
        """
        Paint cell.

        Args:
            painter (QPainter): painter.
            option (QStyleOptionViewItem) option.
            index (QModelIndex): index.
        """
        if index.isValid():
            value = index.model().data(index, Qt.DisplayRole)
            if value:
                option.displayAlignment = Qt.AlignHCenter | Qt.AlignVCenter
                self.parent().style().drawControl(QStyle.CE_ItemViewItem, option, painter, self.parent())


# ==================================================
class ComboDelegate(Delegate):
    # ==================================================
    def __init__(self, parent):
        """
        Create delegate for Combo.

        Args:
            parent (QWidget): parent.
        """
        super().__init__(parent)

    # ==================================================
    def createEditor(self, parent, option, index):
        """
        Create combobox.

        Args:
            parent (QWidget): parent.
            option (QStyleOptionViewItem): option.
            index (QModelIndex): index.

        Returns:
            - (Combo) -- combo widget.
        """
        c = index.column()
        dd = index.model().column_default[c]
        opt = index.model().column_option[c]
        combo = Combo(parent, opt, dd)
        combo.setStyleSheet("margin: 5px 10px 5px 10px; padding: 0px 5px 0px 5px;")
        combo.currentTextChanged.connect(lambda data: index.model().setData(index, data))
        return combo


# ==================================================
class ColorDelegate(Delegate):
    # ==================================================
    def __init__(self, parent):
        """
        Create delegate for ColorSelector.

        Args:
            parent (QWidget): parent.
        """
        super().__init__(parent)

    # ==================================================
    def createEditor(self, parent, option, index):
        """
        Create color selector.

        Args:
            parent (QWidget): parent.
            option (QStyleOptionViewItem): option.
            index (QModelIndex): index.

        Returns:
            - (ColorSelector) -- color selector widget.
        """
        c = index.column()
        dd = index.model().column_default[c]
        tp = index.model().column_type[c]
        color_selector = ColorSelector(parent, dd, tp)
        color_selector.setStyleSheet("margin: 5px 10px 5px 10px; padding: 0px 5px 0px 5px;")
        color_selector.currentTextChanged.connect(lambda data: index.model().setData(index, data))

        return color_selector


# ==================================================
class EditorDelegate(Delegate):
    # ==================================================
    def __init__(self, parent):
        """
        Create delegate for Editor.

        Args:
            parent (QWidget): parent.
        """
        super().__init__(parent)

    # ==================================================
    def createEditor(self, parent, option, index):
        """
        Create editor.

        Args:
            parent (QWidget): parent.
            option (QStyleOptionViewItem): option.
            index (QModelIndex): index.

        Returns:
            - (Editor) -- editor widget.
        """
        model = index.model()
        c = index.column()
        dd = model.column_default[c]
        opt = model.column_option[c]
        tp = model.column_type[c]
        color = model.parent()._preference["latex"]["color"]
        size = model.parent()._preference["latex"]["size"]
        dpi = model.parent()._preference["latex"]["dpi"]
        editor = Editor(parent, dd, (tp, opt), color, size, dpi)

        editor.returnPressed.connect(lambda data: index.model().setData(index, data))
        return editor



"""
Utility for parsing material files.

This module contains utilities for parsing.
"""

import numpy as np
import warnings
import copy
from pymatgen.core import Structure
from pymatgen.core.periodic_table import Element
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.analysis.local_env import MinimumDistanceNN
from pymatgen.io.cif import CifParser
from qtdraw.core.pyvista_widget_setting import DIGIT, default_preference, default_status
from qtdraw.parser.element import element_color
from qtdraw.core.pyvista_widget_setting import widget_detail as detail
from qtdraw.parser.vesta import parse_vesta, create_structure_vesta
from qtdraw.parser.data_group import _shift_vectors, _data_no_space_group
from qtdraw import __version__


# ==================================================
def convert_to_origin_standard(structure, sg_no):
    """
    Convert the orgin for choice 2.

    Args:
        structure (Structure): structure.
        sg_no (int): space group number.
    """
    shift_vector = np.array(_shift_vectors.get(sg_no, [0.0, 0.0, 0.0]))

    for site in structure.sites:
        site.frac_coords += shift_vector
        site.frac_coords %= 1.0  # fractional coordinate -> [0,1].


# ==================================================
def get_model_cell(graph):
    """
    Get model and cell.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.

    Returns:
        - (str) -- name of model.
        - (dict) -- unit-cell info.
    """
    name = graph.structure.composition.reduced_formula

    lat = graph.structure.lattice
    cell = {
        "a": np.round(lat.a, DIGIT),
        "b": np.round(lat.b, DIGIT),
        "c": np.round(lat.c, DIGIT),
        "alpha": np.round(lat.alpha, DIGIT),
        "beta": np.round(lat.beta, DIGIT),
        "gamma": np.round(lat.gamma, DIGIT),
    }

    return name, cell


# ==================================================
def get_site_info(graph):
    """
    Get site information.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.

    Returns:
        - (list) -- site_info. (name, label, element, frac_coords, radius).
    """
    structure = graph.structure
    eq_sites = structure.equivalent_sites

    # grouping equivalent sites.
    dsites = {}
    for es in eq_sites:
        el = es[0].specie
        if type(el) != type(Element):
            el = str(el).replace("0+", "")
        else:
            el = el.symbol
        dsites[el] = dsites.get(el, []) + [es]

    site_info = []
    for name, v in dsites.items():
        for el_no, sl in enumerate(v):
            el_name = name + str(el_no + 1)
            for s_no, s in enumerate(sl):
                s_name = el_name + "_" + str(s_no + 1)
                frac_coords = s.frac_coords.round(DIGIT)
                radius = float(s.specie.atomic_radius)
                site_info.append((el_name, s_name, name, frac_coords, radius))

    return site_info


# ==================================================
def get_bond_info(graph, site_info):
    """
    Get bond information.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.
        site_info (list): site info.

    Returns:
        - (list) -- bond_info. (name, label, center, vector, tail_element, head_element).
    """
    adjacency = graph.as_dict()["graphs"]["adjacency"]

    dbonds = {}
    for tail, tail_adjacency in zip(site_info, adjacency):
        t_name, _, t_element, t_pos, _ = tail
        for head_info in tail_adjacency:
            h_name, _, h_element, h_pos, _ = site_info[head_info["id"]]
            h_pos = np.array(head_info["to_jimage"]) + h_pos
            center = ((t_pos + h_pos) / 2).round(DIGIT)
            vector = (h_pos - t_pos).round(DIGIT)
            key = t_name + "-" + h_name
            dbonds[key] = dbonds.get(key, []) + [(center, vector, t_element, h_element)]

    bond_info = []
    for name, v in dbonds.items():
        for no, (center, vector, t_element, h_element) in enumerate(v):
            label = name + "_" + str(no + 1)
            bond_info.append((name, label, center, vector, t_element, h_element))

    return bond_info


# ==================================================
def draw_site_bond(widget, name, site_info, bond_info):
    """
    Draw site and bond.

    Args:
        widget (PyVistaWidget): widget.
        name (str): name of model.
        site_info (list): site info.
        bond_info (list): bond info.
    """
    site_scale = detail["site_scale"]
    bond_scale = detail["bond_scale"]
    color_scheme = widget._preference["general"]["color_scheme"]
    default_color = "silver"

    # draw sites.
    widget._data["site"].block_update_widget(True)
    for name, label, element, frac_coords, radius in site_info:
        color = element_color[color_scheme].get(element, default_color)
        widget.add_site(position=str(frac_coords.tolist()), name=name, label=label, color=color, size=site_scale * radius)
    widget._data["site"].block_update_widget(False)

    # draw bonds.
    widget._data["bond"].block_update_widget(True)
    for name, label, center, vector, tail_element, head_element in bond_info:
        tail_color = element_color[color_scheme].get(tail_element, default_color)
        head_color = element_color[color_scheme].get(head_element, default_color)
        widget.add_bond(
            position=str(center.tolist()),
            direction=str(vector.tolist()),
            color=tail_color,
            color2=head_color,
            name=name,
            label=label,
            width=bond_scale,
        )
    widget._data["bond"].block_update_widget(False)


# ==================================================
def parse_material(filename):
    """
    Parse material file.

    Args:
        filename (str): filename.

    Returns:
        - (dict) -- data for PyVistaWidget.
        - (list) -- site info. to draw.
        - (list) -- bond info. to draw.
        - (Structure) -- symmetrized structure.
    """
    warnings.filterwarnings("ignore", category=UserWarning)

    if filename.endswith(".vesta"):
        vesta_dict = parse_vesta(filename)
        structure = create_structure_vesta(vesta_dict)
    elif filename.endswith(".cif"):
        parser = CifParser(filename)
        structure = parser.get_structures(primitive=False, symmetrized=False)[0]
    elif filename.endswith(".xsf"):
        structure = Structure.from_file(filename)
        sga = SpacegroupAnalyzer(structure)
        structure = sga.get_refined_structure()
    else:
        structure = Structure.from_file(filename)

    sga = SpacegroupAnalyzer(structure)
    symmetrized = sga.get_symmetrized_structure()
    sg_no = sga.get_space_group_number()
    convert_to_origin_standard(symmetrized, sg_no)
    env = MinimumDistanceNN()
    graph = StructureGraph.with_local_env_strategy(symmetrized, env)

    name, cell = get_model_cell(graph)
    crystal = sga.get_crystal_system()

    group_name = _data_no_space_group[sg_no]
    multipie = {"group": {"group": group_name}}

    status = copy.deepcopy(default_status)
    status.update({"model": name, "crystal": crystal, "cell": cell, "clip": False, "multipie": multipie})
    preference = copy.deepcopy(default_preference)
    preference["axis"]["label"] = "[a,b,c]"
    preference["label"]["default_check"] = False
    all_data = {"version": __version__, "status": status, "preference": preference}

    site_info = get_site_info(graph)
    bond_info = get_bond_info(graph, site_info)

    return all_data, site_info, bond_info, symmetrized



"""
Multipie plugin.

This module provides MultiPie plugin.
"""

import logging
import numpy as np
import sympy as sp
from gcoreutils.nsarray import NSArray
from multipie import __version__, get_binary
from multipie.const import __def_dict__
from multipie.tag.tag_group import TagGroup
from multipie.group.point_group import PointGroup
from multipie.group.space_group import SpaceGroup
from qtdraw.multipie.plugin_multipie_setting import plugin_detail as detail
from qtdraw.multipie.plugin_multipie_setting import default_status
from qtdraw.multipie.dialog_multipie import MultiPieDialog
from qtdraw.multipie.util import check_get_site_bond, check_get_site, check_get_bond, combined_format, create_samb_object


# ==================================================
def _mapping_str(mp):
    """
    Convert mapping No. to those start from one.

    Args:
        mp (list): mapping.

    Returns:
        - (list) -- converted mapping.
    """
    m = [i - 1 if i < 0 else i + 1 for i in mp]
    return str(m).replace(" ", "")


# ==================================================
class MultiPiePlugin:
    # ==================================================
    def __init__(self, parent):
        """
        MultiPie plugin.

        Args:
            parent (QWidget): parent.
        """
        self.parent = parent
        self._pvw = parent.pyvista_widget
        self._core = get_binary()

        multipie_status = default_status
        multipie_status["plus"] = {}
        for key, value in self._pvw._status["multipie"].items():
            if type(value) == dict:
                multipie_status[key].update(value)
        self._pvw._status["multipie"] = multipie_status
        self._pvw._status["multipie"]["version"] = self.version

        self._pvw.update_preference("label", "default_check", detail["general"]["label"])

        self.set_group()

        self.dialog = MultiPieDialog(self)
        self.dialog.show()

    # ==================================================
    @property
    def version(self):
        """
        Version.
        """
        return __version__

    # ==================================================
    @property
    def group(self):
        """
        Group status dict.
        """
        return self._pvw._status["multipie"]["group"]

    # ==================================================
    @property
    def obj(self):
        """
        Object status dict.
        """
        return self._pvw._status["multipie"]["object"]

    # ==================================================
    @property
    def basis(self):
        """
        Basis status dict.
        """
        return self._pvw._status["multipie"]["basis"]

    # ==================================================
    @property
    def plus(self):
        """
        Plus status dict.
        """
        return self._pvw._status["multipie"]["plus"]

    # ==================================================
    def counter(self, name):
        """
        Counter number and increment.

        Args:
            name (str): name of counter.

        Returns:
            - (int) -- counter.
        """
        n = self._pvw._status["multipie"]["counter"][name]
        self._pvw._status["multipie"]["counter"][name] = n + 1
        return n

    # ==================================================
    def clear_counter(self):
        """
        Clear counter.
        """
        for key in self._pvw._status["multipie"]["counter"].keys():
            self._pvw._status["multipie"]["counter"][key] = 0

    # ==================================================
    def set_group(self):
        """
        Set group and plus dict.
        """
        tag = TagGroup(self.group["group"])

        pg = tag.is_point_group()
        self.plus["point_group"] = pg
        if pg:
            self._group = PointGroup(tag, self._core)
            self._point_group = self._group
            self.plus["n_pset"] = 1
            self.plus["pset"] = "{[0,0,0]}"
        else:
            self._group = SpaceGroup(tag, self._core)
            self._point_group = self._group.pg
            self.plus["n_pset"] = len(self._group.symmetry_operation.plus_set)
            self.plus["pset"] = str(self._group.symmetry_operation.plus_set)

        self.plus["crystal"] = self._group.tag.crystal
        self.plus["wyckoff"] = list(map(str, self._point_group.wyckoff.keys()))[::-1]
        self.plus["irrep"] = list(map(str, self._point_group.character.irrep_list))
        self.plus["n_op"] = len(self._group.symmetry_operation)

        # for old QtDraw version.
        if type(self.group["irrep1"]) == int:
            self.group["irrep1"] = self.plus["irrep"][self.group["irrep1"]]
        if type(self.group["irrep2"]) == int:
            self.group["irrep2"] = self.plus["irrep"][self.group["irrep2"]]
        if type(self.group["irrep"]) == int:
            self.group["irrep"] = self.plus["irrep"][self.group["irrep"]]
        if type(self.group["vc_wyckoff"]) == int:
            self.group["vc_wyckoff"] = self.plus["wyckoff"][self.group["vc_wyckoff"]]

        # empty basis.
        self.plus["site_cluster"] = ""
        self.plus["site_z_samb"] = {}
        self.plus["site_c_samb"] = {}

        self.plus["bond_cluster"] = ""
        self.plus["bond_z_samb"] = {}
        self.plus["bond_c_samb"] = {}

        self.plus["vector_cluster"] = ""
        self.plus["vector_z_samb"] = {}
        self.plus["vector_c_samb"] = {}

        self.plus["orbital_cluster"] = ""
        self.plus["orbital_z_samb"] = {}
        self.plus["orbital_c_samb"] = {}

    # ==================================================
    def get_product_decomp(self, irrep1, irrep2, irrep):
        """
        Set product decomposition.

        Args:
            irrep1 (str): irrep1 for symmetric product.
            irrep2 (str): irrep2 for symmetric product.
            irrep (str): irerp for anti symmetric product.

        Returns:
            - (str) -- symmetric decomposition.
            - (str) -- anti symmetric decomposition.
        """

        def remove_latex(s):
            s = (
                s.replace("_{", "")
                .replace("}", "")
                .replace("^{", "")
                .replace(r"\prime", "'")
                .replace("(", "")
                .replace(")", "")
                .replace("0", "-")
            )
            return s

        s = self._point_group.character.symmetric_product_decomposition((irrep1, irrep2), ret_ex=True)
        a = self._point_group.character.anti_symmetric_product_decomposition(irrep, ret_ex=True)

        s = remove_latex(str(s))
        a = remove_latex(str(a))

        return s, a

    # ==================================================
    def get_harmonics_irrep(self):
        """
        Get harmonics irrep.

        Returns:
            - (list) -- harmonics name list.
        """
        h_type = self.obj["harmonics_type"]
        rank = self.obj["harmonics_rank"]
        h_type1 = h_type.replace("T", "Q").replace("M", "G")

        hs = self._point_group.harmonics.select(rank=rank, head=h_type1)
        lst = [h_type + str(i)[2:] for i in hs]

        return lst

    # ==================================================
    def get_harmonics(self):
        """
        Get harmonics.

        Returns:
            - (str) -- harmonics expression (in LaTeX).
        """
        h_type = self.obj["harmonics_type"]
        rank = self.obj["harmonics_rank"]
        irrep = self.obj["harmonics_irrep"]
        h_type = h_type.replace("T", "Q").replace("M", "G")
        hs = self._point_group.harmonics.select(rank=rank, head=h_type)
        h = hs[irrep]
        ex = h.expression(v=NSArray.vector3d("Q"))
        if self.obj["harmonics_latex"]:
            ex = ex.latex()
        else:
            ex = str(ex)

        return ex

    # ==================================================
    def find_wyckoff(self):
        """
        Find Wyckoff position.

        Returns:
            - (str) -- Wyckoff position.
            - (str) -- site symmetry.
        """
        wyckoff = self.obj["wyckoff"]
        r_site = check_get_site_bond(wyckoff, ret_site=True)
        if r_site is None:
            logging.error(f"cannot find Wyckoff position, {wyckoff}")
            return None, None

        wp = self._group.find_wyckoff_position(r_site)
        sym = self._group.wyckoff.site_symmetry(wp)
        wp = str(wp)

        return wp, sym

    # ==================================================
    def add_equivalent_site(self, site, scale=1.0, color=None):
        """
        Add equivalent sites.

        Args:
            site (str): representative site.
            scale (float, optional): size scale.
            color (str, optional): color.

        Returns:
            - (str) -- representative site.
        """
        n_pset = self.plus["n_pset"]
        r_site = check_get_site(site)
        if r_site is None:
            logging.error(f"format error, {site}")
            return None

        if self.plus["point_group"]:
            site = self._group.site_mapping(r_site)
        else:
            site = self._group.site_mapping(r_site, plus_set=True)

        count = self.counter("site") + 1
        name0 = f"S{count:02}"
        if color is None:
            color = detail["general"]["site_color"]
        size = detail["object"]["site_size"]
        primitive_num = len(site) // n_pset

        for no, (s, mp) in enumerate(site.items()):
            s = NSArray(s)
            mp = _mapping_str(mp)
            idx = no % primitive_num
            pset = no // primitive_num
            name = name0
            if n_pset > 1:
                name += f"({pset+1})"
            label = f"s{idx+1:02}:{mp}"
            self._pvw.add_site(position=s.value(), size=size * scale, color=color, name=name, label=label)

        return str(r_site)

    # ==================================================
    def add_equivalent_bond(self, bond, scale=1.0, color=None, color2=None):
        """
        Add equivalent bonds.

        Args:
            bond (str): representative bond.
            scale (float, optional): size scale.
            color (str, optional): color.
            color2 (str, optional): color2.

        Returns:
            - (str) -- representative bond.
        """
        n_pset = self.plus["n_pset"]
        r_bond = check_get_bond(bond)
        if r_bond is None:
            logging.error(f"format error, {bond}")
            return None

        if self.plus["point_group"]:
            bond, nondirectional = self._group.bond_mapping(r_bond)
        else:
            bond, nondirectional = self._group.bond_mapping(r_bond, plus_set=True)

        count = self.counter("bond") + 1
        name0 = f"B{count:02}"
        if color is None:
            color1 = detail["general"]["bond_color1"]
        else:
            color1 = color
        if nondirectional:
            color2 = color1
        else:
            if color2 is None:
                color2 = detail["general"]["bond_color2"]
        primitive_num = len(bond) // n_pset
        width = detail["object"]["bond_width"]

        for no, (b, mp) in enumerate(bond.items()):
            b = NSArray(b)
            v, c = b.convert_bond("bond")
            mp = _mapping_str(mp)
            idx = no % primitive_num
            pset = no // primitive_num
            name = name0
            if n_pset > 1:
                name += f"({pset+1})"
            label = f"b{idx+1:02}:{mp}"
            self._pvw.add_bond(
                position=c.value(), direction=v.value(), color=color1, color2=color2, width=width * scale, name=name, label=label
            )

        return str(r_bond)

    # ==================================================
    def add_vector_equivalent_site(self, v_type, pos, scale=1.0):
        """
        Add vectors at equivalent sites.

        Args:
            v_type (str): vector type.
            pos (str): vector # position.
            scale (float, optional): length scale.

        Returns:
            - (str) -- vector # position.
        """
        n_pset = self.plus["n_pset"]

        txt = pos.split("#")
        if len(txt) != 2:
            logging.error(f"format error, {pos}")
            return None
        vector, r_site_bond = txt
        vector = check_get_site(vector)
        if vector is None:
            logging.error(f"format error, {pos}")
            return None
        r_site = check_get_site_bond(r_site_bond, ret_site=True)
        if r_site is None:
            logging.error(f"format error, {pos}")
            return None

        if self.plus["point_group"]:
            site = self._group.site_mapping(r_site)
        else:
            site = self._group.site_mapping(r_site, plus_set=True)
        site = NSArray.from_str(site.keys())

        count = self.counter("vector") + 1
        name0 = f"V{count:02}"
        color = detail["general"]["vector_color_" + v_type]
        primitive_num = len(site) // n_pset
        length = detail["object"]["vector_length"]

        for no, s in enumerate(site):
            idx = no % primitive_num
            pset = no // primitive_num
            name = name0
            if n_pset > 1:
                name += f"({pset+1})"
            label = f"v{idx+1:02}"
            self._pvw.add_vector(position=s.value(), direction=vector, length=length * scale, color=color, name=name, label=label)

        return str(vector) + "#" + str(r_site)

    # ==================================================
    def add_orbital_equivalent_site(self, o_type, pos, scale=1.0):
        """
        Add orbitals at equivalent sites.

        Args:
            o_type (str): orbital type.
            pos (str): orbital # position.
            scale (float, optional): size scale.

        Returns:
            - (str) -- orbital # position.
        """
        n_pset = self.plus["n_pset"]

        txt = pos.split("#")
        if len(txt) != 2:
            logging.error(f"format error, {pos}")
            return None
        orbital, r_site_bond = txt
        r_site = check_get_site_bond(r_site_bond, ret_site=True)
        if r_site is None:
            logging.error(f"format error, {pos}")
            return None

        if self.plus["point_group"]:
            site = self._group.site_mapping(r_site)
        else:
            site = self._group.site_mapping(r_site, plus_set=True)
        site = NSArray.from_str(site.keys())

        count = self.counter("orbital") + 1
        name0 = f"O{count:02}"
        color = detail["general"]["orbital_color_" + o_type]
        primitive_num = len(site) // n_pset
        size = detail["object"]["orbital_size"]

        for no, s in enumerate(site):
            idx = no % primitive_num
            pset = no // primitive_num
            name = name0
            if n_pset > 1:
                name += f"({pset+1})"
            label = f"o{idx+1:02}"
            self._pvw.add_orbital(position=s.value(), shape=orbital, size=size * scale, color=color, name=name, label=label)

        return str(orbital) + "#" + str(r_site)

    # ==================================================
    def create_combined(self, site_bond, rank, head, ret_bond=False):
        """
        Create combined SAMB.

        Args:
            site_bond (str): site or bond.
            rank (str): harmonics rank.
            head (str): harmonics type.
            ret_bond (bool, optional): return bond ?

        Returns:
            - (dict) -- cluster SAMB.
            - (NSArray) -- site or bond.
            - (dict) -- combined SAMB.
        """
        t_rev = {"Q": "Q", "G": "G", "T": "Q", "M": "G"}

        if site_bond.style == "vector":
            c_samb, site = self._group.site_cluster_samb(site_bond)
        else:
            c_samb, bond = self._group.bond_cluster_samb(site_bond)
            if ret_bond:
                site = bond
            else:
                site = bond.convert_bond("bond")[1]

        x_tag = self._point_group.harmonics.key_list().select(rank=int(rank), head=t_rev[head])
        if head in ["T", "M"]:
            x_tag = [tag.reverse_t_type() for tag in x_tag]
        y_tag = list(c_samb.keys())

        z_samb_all = self._group.z_samb(x_tag, y_tag)
        z_samb = {"Q": [], "G": [], "T": [], "M": []}
        for tag, c in z_samb_all.items():
            tag_str = combined_format(tag)
            z_samb[tag[0].head].append((tag_str, c))
        for k in z_samb.keys():
            z_samb[k] = list(sorted(z_samb[k], key=lambda i: i[0]))

        return c_samb, site, z_samb

    # ==================================================
    def gen_site_samb(self, pos):
        """
        Generate site cluster SAMB.

        Args:
            pos (str): site.

        Returns:
            - (str) -- position.
        """
        pos = check_get_site(pos)
        if pos is None:
            logging.error(f"format error, {pos}")
            self.plus["site_cluster"] = ""
            self.plus["site_z_samb"] = {}
            self.plus["site_c_samb"] = {}
            return ""

        combined_info = self.create_combined(pos, 0, "Q")
        c_samb, cluster, z_samb = combined_info
        self.plus["site_cluster"] = cluster
        self.plus["site_z_samb"] = z_samb
        self.plus["site_c_samb"] = c_samb
        return str(pos)

    # ==================================================
    def gen_bond_samb(self, pos):
        """
        Generate bond cluster SAMB.

        Args:
            pos (str): bond.

        Returns:
            - (str) -- position.
        """
        pos = check_get_bond(pos)
        if pos is None:
            logging.error(f"format error, {pos}")
            self.plus["bond_cluster"] = ""
            self.plus["bond_z_samb"] = {}
            self.plus["bond_c_samb"] = {}
            return ""

        combined_info = self.create_combined(pos, 0, "Q", ret_bond=True)
        c_samb, cluster, z_samb = combined_info
        self.plus["bond_cluster"] = cluster
        self.plus["bond_z_samb"] = z_samb
        self.plus["bond_c_samb"] = c_samb
        return str(pos)

    # ==================================================
    def gen_vector_samb(self, pos, v_type):
        """
        Generate vector cluster SAMB.

        Args:
            pos (str): vector # position.
            v_type (str): vector type.

        Returns:
            - (str) -- position.
        """
        pos = check_get_site_bond(pos)
        if pos is None:
            logging.error(f"format error, {pos}")
            self.plus["vector_cluster"] = ""
            self.plus["vector_z_samb"] = {}
            self.plus["vector_c_samb"] = {}
            return ""

        combined_info = self.create_combined(pos, 1, v_type)
        c_samb, cluster, z_samb = combined_info
        self.plus["vector_cluster"] = cluster
        self.plus["vector_z_samb"] = z_samb
        self.plus["vector_c_samb"] = c_samb
        return str(pos)

    # ==================================================
    def gen_orbital_samb(self, pos, o_type, rank):
        """
        Generate orbital cluster SAMB.

        Args:
            pos (str): orbital # position.
            o_type (str): orbital type.
            rank (str): orbital rank.

        Returns:
            - (str) -- position.
        """
        pos = check_get_site_bond(pos)
        if pos is None:
            logging.error(f"format error, {pos}")
            self.plus["orbital_cluster"] = ""
            self.plus["orbital_z_samb"] = {}
            self.plus["orbital_c_samb"] = {}
            return ""

        combined_info = self.create_combined(pos, rank, o_type)
        c_samb, cluster, z_samb = combined_info
        self.plus["orbital_cluster"] = cluster
        self.plus["orbital_z_samb"] = z_samb
        self.plus["orbital_c_samb"] = c_samb
        return str(pos)

    # ==================================================
    def gen_hopping_samb(self, pos):
        """
        Generate hopping SAMB.

        Args:
            pos (str): bond.

        Returns:
            - (str) -- position.
        """
        pos = check_get_bond(pos)
        if pos is None:
            logging.error(f"format error, {pos}")
            return ""
        return str(pos)

    # ==================================================
    def create_hopping_direction(self, bond):
        """
        Create normal bond direction.

        Args:
            bond (str): bond.

        Returns:
            - (NSArray) -- a set of bonds (no plus_set).
        """
        pg = self.plus["point_group"]
        bond = NSArray(bond)

        site = bond.convert_bond("bond_th")[0]
        if pg:
            sites = list(self._group.site_mapping(site).keys())
        else:
            sites = list(self._group.site_mapping(site, plus_set=True).keys())
        bonds = NSArray.from_str(list(self._group.bond_mapping(bond)[0].keys()))
        new_bonds = []
        for i in range(len(bonds)):
            t = bonds[i].convert_bond("bond_th")[0]
            if not pg:
                t = t.shift()
            if str(t) not in sites:
                new_bonds.append(str(bonds[i].reverse_direction()))
            else:
                new_bonds.append(str(bonds[i]))

        new_bonds = NSArray.from_str(new_bonds)

        return new_bonds

    # ==================================================
    def add_site_samb(self, site, obj, label, scale=1.0):
        """
        Add site cluster SAMB.

        Args:
            site (NSArray): equivalent sites.
            obj (NSArray): SAMB weight.
            label (str): label.
            scale (float, optional): size scale.
        """
        pset = NSArray(self.plus["pset"])
        pg = self.plus["point_group"]

        color = []
        for w in obj:
            if w > 0:
                c = "salmon"
            elif w < 0:
                c = "aqua"
            else:
                c = "silver"
            color.append(c)

        obj /= np.abs(obj).max()
        n_pset = len(pset)
        size = detail["samb"]["site_size"]
        if scale is None or type(scale) == bool:
            scale = detail["samb"]["site_scale"]

        count = self.counter("site_samb") + 1
        name = f"Z_{count:03}"
        for no, p in enumerate(pset):
            name1 = name
            if n_pset != 1:
                name1 += f"({no+1})"
            for s, w, cl in zip(site, obj, color):
                if not pg:
                    s = (s + p).shift()
                if cl == "silver":
                    w = 1
                self._pvw.add_site(position=s.value(), size=size * abs(w) * scale, color=cl, name=name1, label=label)

    # ==================================================
    def add_bond_samb(self, bond, obj, label, z_type, scale=1.0):
        """
        Add bond cluster SAMB.

        Args:
            bond (NSArray): equivalent bonds.
            obj (NSArray): SAMB weight.
            label (str): label.
            z_type (str): SAMB type.
            scale (float, optional): width scale.
        """
        pset = NSArray(self.plus["pset"])
        pg = self.plus["point_group"]
        A = NSArray(self._pvw.A_matrix, "matrix")

        color = []
        if z_type == "Q":
            for w in obj:
                if w > 0:
                    c = "salmon"
                elif w < 0:
                    c = "aqua"
                else:
                    c = "silver"
                color.append(c)
        else:
            for w in obj:
                if w == 0:
                    c = "silver"
                else:
                    c = "salmon"
                color.append(c)

        obj /= np.abs(obj).max()
        n_pset = len(pset)
        width = detail["samb"]["bond_width"]
        if scale is None or type(scale) == bool:
            scale = detail["samb"]["bond_scale"]

        count = self.counter("bond_samb") + 1
        name = f"Z_{count:02}"
        if z_type == "Q":
            for no, p in enumerate(pset):
                name1 = name
                if n_pset != 1:
                    name1 += f"({no+1})"
                for s, w, cl in zip(bond, obj, color):
                    v, c = s.convert_bond("bond")
                    if not pg:
                        c = (c + p).shift()
                    if cl == "silver":
                        w = 1
                    self._pvw.add_bond(
                        position=c.value(),
                        direction=v.value(),
                        color=cl,
                        color2=cl,
                        width=width * abs(w) * scale,
                        name=name1,
                        label=label,
                    )
        else:
            for no, p in enumerate(pset):
                name1 = name
                if n_pset != 1:
                    name1 += f"({no+1})"
                for s, w, cl in zip(bond, obj, color):
                    v, c = s.convert_bond("bond")
                    if not pg:
                        c = (c + p).shift()
                    if cl == "silver":
                        w = 1
                        self._pvw.add_bond(
                            position=c.value(),
                            direction=v.value(),
                            color=cl,
                            color2=cl,
                            width=width * abs(w) * scale,
                            name=name1,
                            label=label,
                        )
                    else:
                        v = v.transform(A)
                        if w < 0:
                            v = -v
                        norm = v.norm() * scale
                        self._pvw.add_vector(
                            position=c.value(),
                            direction=v.value(),
                            color=cl,
                            width=width * abs(w) * scale,
                            length=norm,
                            offset=-0.5,
                            name=name1,
                            label=label,
                        )

    # ==================================================
    def add_vector_samb(self, site, obj, label, z_type, v, scale=1.0):
        """
        Add vector cluster SAMB.

        Args:
            site (NSArray): equivalent sites.
            obj (NSArray): SAMB weight.
            label (str): label.
            z_type (str): SAMB type.
            v (NSArray): vector variable.
            scale (float, optional): length scale.
        """
        pset = NSArray(self.plus["pset"])
        pg = self.plus["point_group"]

        rep = {
            v[0]: sp.Matrix([1, 0, 0]),
            v[1]: sp.Matrix([0, 1, 0]),
            v[2]: sp.Matrix([0, 0, 1]),
        }
        color = detail["general"]["vector_color_" + z_type]
        count = self.counter("vector_samb") + 1
        name = f"Z_{count:03}"
        n_pset = len(pset)
        width = detail["samb"]["vector_width"]

        for no, p in enumerate(pset):
            name1 = name
            if n_pset != 1:
                name1 += f"({no+1})"
            for s, c in zip(site, obj):
                if not pg:
                    s = (s + p).shift()
                if c != 0:
                    c = str(c.subs(rep).T[:])
                    c = NSArray(c)
                    d = c.norm()
                    self._pvw.add_vector(
                        position=s.value(),
                        direction=c.value(),
                        width=width,
                        length=d.value() * scale,
                        color=color,
                        name=name1,
                        label=label,
                    )

    # ==================================================
    def add_orbital_samb(self, site, obj, label, z_type, scale=1.0):
        """
        Add orbital cluster SAMB.

        Args:
            site (NSArray): equivalent sites.
            obj (NSArray): SAMB weight.
            label (str): label.
            z_type (str): SAMB type.
            scale (float, optional): size scale.
        """
        pset = NSArray(self.plus["pset"])
        pg = self.plus["point_group"]

        color = detail["general"]["orbital_color_" + z_type]
        count = self.counter("orbital_samb") + 1
        name = f"Z_{count:03}"
        n_pset = len(pset)
        size = detail["samb"]["orbital_size"]

        for no, p in enumerate(pset):
            name1 = name
            if n_pset != 1:
                name1 += f"({no+1})"
            for s, orb in zip(site, obj):
                if not pg:
                    s = (s + p).shift()
                self._pvw.add_orbital(
                    position=s.value(),
                    shape=orb,
                    size=size * scale,
                    color=color,
                    name=name1,
                    label=label,
                )

    # ==================================================
    def add_vector_modulation(self, obj, igrid, head, v):
        """
        Add vector SAMB modulation.

        Args:
            obj (list): SAMB weight at (cell,pset).
            igrid (NSArray): cell grid.
            head (str): multipole type.
            v (NSArray): vector variable.
        """
        rep = {
            v[0]: sp.Matrix([1, 0, 0]),
            v[1]: sp.Matrix([0, 1, 0]),
            v[2]: sp.Matrix([0, 0, 1]),
        }
        color = detail["general"]["vector_color_" + head]
        pset = NSArray(self.plus["pset"])
        n_pset = self.plus["n_pset"]
        cluster = self.plus["vector_cluster"]

        count = self.counter("vector") + 1
        name = f"mod{count:02}"
        for i_no, i in enumerate(igrid):
            for p_no, p in enumerate(pset):
                label = ""
                if n_pset != 1:
                    label = f"({p_no+1})"
                for no, (s, c) in enumerate(zip(cluster, obj[i_no][p_no])):
                    s = (s + p).shift()
                    label1 = f"s{no+1}" + label
                    if c != 0:
                        c = str(c.subs(rep).T[:])
                        c = NSArray(c)
                        d = c.norm().value()
                        self._pvw.add_vector(
                            position=s.value() + np.array(i), direction=c.value(), length=d, color=color, name=name, label=label1
                        )

    # ==================================================
    def add_orbital_modulation(self, obj, igrid, head):
        """
        Add orbitl SAMB modulation.

        Args:
            obj (list): SAMB weight at (cell,pset).
            igrid (NSArray): cell grid.
            head (str): multipole type.
        """
        color = detail["general"]["orbital_color_" + head]
        pset = NSArray(self.plus["pset"])
        n_pset = self.plus["n_pset"]
        cluster = self.plus["orbital_cluster"]
        size = detail["samb"]["orbital_mod"]

        count = self.counter("orbital") + 1
        name = f"mod{count:02}"
        for i_no, i in enumerate(igrid):
            for p_no, p in enumerate(pset):
                label = ""
                if n_pset != 1:
                    label += f"({p_no+1})"
                for no, (s, orb) in enumerate(zip(cluster, obj[i_no][p_no])):
                    s = (s + p).shift()
                    label1 = f"s{no+1}" + label
                    self._pvw.add_orbital(
                        position=s.value() + np.array(i),
                        shape=str(orb),
                        surface="",
                        size=size,
                        color=color,
                        name=name,
                        label=label1,
                    )

    # ==================================================
    def add_hopping_samb(self, bond, label, scale=1.0):
        """
        Add normal hopping direction.

        Args:
            bond (NSArray): equivalent bonds.
            label (str): label.
            scale (float, optional): length scale.
        """
        pset = NSArray(self.plus["pset"])
        pg = self.plus["point_group"]

        color = "salmon"
        n_pset = len(pset)

        count = self.counter("hopping") + 1
        name = f"Z_{count:02}"
        A = NSArray(self._pvw.A_matrix, "matrix")
        for no, p in enumerate(pset):
            name1 = name
            if n_pset != 1:
                name1 += f"({no+1:02})"
            for s in bond:
                v, c = s.convert_bond("bond")
                if not pg:
                    c = (c + p).shift()
                v = v.transform(A)
                norm = v.norm() * scale
                self._pvw.add_vector(
                    position=c.value(),
                    direction=v.value(),
                    color=color,
                    length=norm * scale,
                    offset=-0.5,
                    name=name1,
                    label=label,
                )

    # ==================================================
    def add_harmonics_set(self, head, rank):
        """
        Add harmonics set.

        Args:
            head (str): harmonics type.
            rank (int): harmonics rank.
        """
        color = detail["general"]["orbital_color_" + head]
        count = self.counter("orbital") + 1
        pname = f"{head}{rank:02}({count:02})"
        size = detail["group"]["harmonics_size"]

        pgh = self._point_group.harmonics
        head = head.replace("T", "Q").replace("M", "G")
        hs = pgh.select(rank=rank, head=head)
        n = len(hs)

        pos = NSArray(
            [[np.cos(2 * np.pi * i / n), np.sin(2 * np.pi * i / n), 0.0] for i in range(n)],
            "vector",
            "value",
        )
        for i in range(n):
            self._pvw.add_orbital(
                position=pos[i].value(),
                shape=str(hs[i].expression(v=NSArray.vector3d())),
                size=size,
                color=color,
                name=pname,
                label=f"{i+1:02}",
            )

    # ==================================================
    def add_virtual_cluster(self, wp, bond):
        """
        Add virtual cluster.

        Args:
            wp (str): Wyckoff position.
            bond (str): bond neighbor list.
        """
        _, site = self._group.virtual_cluster_basis(wyckoff=wp)

        pname = wp
        color = detail["general"]["site_color"]
        for i in range(len(site)):
            self._pvw.add_site(
                position=site[i].value(),
                color=color,
                name=pname,
                label=f"{i+1:02}",
            )

        bond = bond.strip("[]")
        bond = list(map(int, bond.split(",")))
        G = NSArray(self._pvw.G_matrix[0:3, 0:3], style="matrix")
        d = NSArray.distance(site, site, G)
        dkey = list(d.keys())
        for i in bond:
            name = f"b{i:02}"
            if i < len(d):
                for idxs in d[dkey[i]]:
                    t, h = site[idxs[0]], site[idxs[1]]
                    c = (t + h) / 2
                    v = h - t
                    self._pvw.add_bond(position=c.value(), direction=v.value(), name=name)

    # ==================================================
    def create_samb_modulation(self, samb_type, v, head, modulation, repeat, offset):
        """
        Create SAMB modulation object.

        Args:
            v (NSArray): vector variable.
            head (str): _description_
            modulation (list): modulation info.
            repeat (list): repeat.
            offset (list): offset.

        Returns:
            - (list) -- [(xyz)-polynomial at each cluster site.] [cell][pset].
            - (list) -- cell grid.
        """
        phase_dict, igrid = self.create_phase_factor(modulation, repeat, offset)
        n_pset = self.plus["n_pset"]

        cluster = self.plus[samb_type + "_cluster"]
        z_samb = self.plus[samb_type + "_z_samb"]
        c_samb = self.plus[samb_type + "_c_samb"]

        obj = NSArray.zeros((len(igrid) * n_pset, len(cluster)), "vector")
        for p_no in range(n_pset):
            for basis, coeff, k, n in modulation:
                z_head = basis[0]
                irrep = int(basis[1:]) - 1
                t_odd = head.replace("M", "T").replace("G", "Q") != z_head.replace("M", "T").replace("G", "Q")
                phase = phase_dict[(k, n, p_no)]
                coeff = NSArray(coeff, fmt="value")
                cluster_obj = create_samb_object(
                    z_samb,
                    cluster,
                    c_samb,
                    z_head,
                    irrep,
                    self._point_group,
                    v,
                    t_odd,
                )
                for i_no in range(len(igrid)):
                    obj[i_no * n_pset + p_no] += coeff * phase[i_no] * cluster_obj

        obj = obj.numpy().reshape((len(igrid), n_pset, len(cluster))).tolist()

        return obj, igrid

    # ==================================================
    def create_phase_factor(self, modulation, repeat, offset):
        """
        Create phase factor.

        Args:
            modulation (list): modulation info. [(basis, coeff, k, n)].
            repeat (list): repeat.
            offset (list): offset.

        Returns:
            - (dict) -- {(k(str),n(int),plus_set no(int)): [phase at each grid(float)]}.
            - (list) -- cell grid.
        """
        pset = NSArray(self.plus["pset"])
        igrid = NSArray.igrid(repeat, offset)

        phase_dict = {}
        for _, _, k, n in modulation:
            kvec = np.array(NSArray(k).tolist(), dtype=float)
            for p_no, p in enumerate(pset):
                lst = []
                for i in igrid:
                    kr = 2.0 * np.pi * kvec @ (i.value() + p.value())
                    phase = np.cos(kr) if n == "cos" else np.sin(kr)
                    lst.append(phase)
                phase_dict[(k, n, p_no)] = lst

        return phase_dict, igrid.numpy().astype(int).tolist()



"""
Two-layer tree model for managing objects.

This module provides a class to manage objects to draw.
The objects are grouped by the first column, "name".
All the objects in the same group can be modified together.
When there are more than two objects in the same group,
a child tree appears and the parent tree is the same as
the first row of the child tree.

The raw data is maintained by dict, and the index model
is just to keep the relation between parent and child,
which is necessary to use other Qt functionalities.
"""

import copy
from PySide6.QtGui import QStandardItemModel, QStandardItem
from PySide6.QtCore import Signal, Qt, QModelIndex, QTimer
from qtdraw.core.pyvista_widget_setting import CUSTOM_WIDGET
from qtdraw.core.pyvista_widget_setting import COLUMN_NAME_ACTOR, COLUMN_LABEL_ACTOR


# ==================================================
class GroupModel(QStandardItemModel):
    updateWidget = Signal()  # update widget when append/remove rows.
    updateData = Signal(str, list, int, QModelIndex)  # name, row_data, role, index.

    # data/check state changed signal for user.
    dataRemoved = Signal(str, list, QModelIndex)  # object_type, row_data, index.
    dataModified = Signal(str, list, QModelIndex)  # object_type, row_data, index.
    checkChanged = Signal(str, list, QModelIndex)  # object_type, row_data, index.

    # role.
    AppendRow = Qt.UserRole + 11
    RemoveRow = Qt.UserRole + 12
    MoveRow = Qt.UserRole + 13

    # ==================================================
    def __init__(self, name, column_info, parent=None):
        """
        Group data model (2 layer parent-child tree model).

        Args:
            name (str): model name.
            column_info (list): {header: (type,option,default)} for each column.
            parent (QWidget, optional): parent.
        """
        super().__init__(parent)
        self._name = name
        self._update_widget = True

        self.setHorizontalHeaderLabels(column_info.keys())
        self.column_type = []
        self.column_option = []
        self.column_default = []
        for c_type, c_option, c_default in column_info.values():
            self.column_type.append(c_type)
            self.column_option.append(c_option)
            self.column_default.append(c_default)

        self.column_widget = [i for i, c in enumerate(self.column_type) if c in CUSTOM_WIDGET]

        self.dataChanged.connect(self.update_check_state)  # to modify bool data in column+1.
        self.updateData.connect(self.emit_update_data)

    # ==================================================
    @property
    def group_name(self):
        """
        Group name.

        Returns:
            - (str) -- group name.
        """
        return self._name

    # ==================================================
    @property
    def header(self):
        """
        Header label.

        Returns:
            - (list) -- header label.
        """
        return [self.headerData(c, Qt.Horizontal) for c in range(self.columnCount())]

    # ==================================================
    def block_update_widget(self, flag):
        """
        Block update widget.

        Args:
            flag (bool): block update ?

        Note:
            - in order to avoid unnecessary update of widget, otherwise it becomes very slow.
        """
        self._update_widget = not flag
        if not flag:
            self.updateWidget.emit()

    # ==================================================
    def is_parent(self, index):
        """
        Is parent index ?

        Args:
            index (QModelIndex): index.

        Returns:
            - (bool) -- parent index ?
        """
        depth = 0
        idx = index
        while idx.isValid():
            idx = idx.parent()
            depth += 1
        return depth == 1

    # ==================================================
    def update_check_state(self, topLeft, bottomRight, roles=[]):
        """
        Update check state data.

        Args:
            topLeft (QModelIndex): top left index.
            bottomRight (QModelIndex): bottom right index.
            roles (list, optional): list of roles.

        Note:
            - set check state (bool) in column+1.
        """
        if not Qt.CheckStateRole in roles:
            return

        for row in range(topLeft.row(), bottomRight.row() + 1):
            col = topLeft.column()
            index = self.index(row, col, topLeft.parent())
            index1 = self.index(row, col + 1, topLeft.parent())
            value = self.data(index, Qt.CheckStateRole)
            state = str(Qt.CheckState(value) == Qt.Checked)
            super().setData(index1, state)

    # ==================================================
    def set_data(self, data):
        """
        Set data from list data.

        Args:
            - (list) -- set data from string.
        """
        for row_data in data:
            self.append_row(row_data)

    # ==================================================
    def set_row_data(self, index, column, data):
        """
        Set row data.

        Args:
            index (QModelIndex): index.
            column (int): column.
            data (str): data.
        """
        self.blockSignals(True)
        super().setData(index.siblingAtColumn(column), data, Qt.EditRole)
        self.blockSignals(False)

    # ==================================================
    def set_check(self, index, column, data):
        """
        Set check state.

        Args:
            index (QModelIndex): index.
            column (int): column.
            data (QCheckState): check state.
        """
        index_c = index.siblingAtColumn(column)
        self.setData(index_c, data, Qt.CheckStateRole)

    # ==================================================
    def tolist(self):
        """
        Convert to list.

        Args:
            raw (bool, optional): raw data ?

        Returns:
            - (list) -- list data.
        """
        root_item = self.invisibleRootItem()
        data = []
        for parent_row in range(root_item.rowCount()):
            item = root_item.child(parent_row)
            if item.hasChildren():
                for row in range(item.rowCount()):
                    row_data = self.get_row_data(item.child(row).index())
                    data.append(row_data)
            else:
                row_data = self.get_row_data(root_item.child(parent_row).index())
                data.append(row_data)

        return data

    # ==================================================
    def emit_update_all(self):
        """
        Emit update for all data.
        """
        name = self.group_name
        root_item = self.invisibleRootItem()
        for parent_row in range(root_item.rowCount()):
            item = root_item.child(parent_row)
            if item.hasChildren():
                for row in range(item.rowCount()):
                    index = item.child(row).index()
                    row_data = self.get_row_data(index)
                    self.dataModified.emit(name, row_data, index)
            else:
                index = root_item.child(parent_row).index()
                row_data = self.get_row_data(index)
                self.dataModified.emit(name, row_data, index)

    # ==================================================
    def find_item(self, text, column=0, child=True):
        """
        Find item.

        Args:
            text (str): str to find.
            column (int, optional): column.
            child (bool, optional): find also for child ?

        Returns:
            - (list) -- found items.
        """
        found = []

        parent_item = self.invisibleRootItem()
        for row in range(parent_item.rowCount()):
            item = parent_item.child(row, column)
            if item and item.data(Qt.EditRole) == text:
                found.append(item)

            if child:
                item0 = parent_item.child(row, 0)
                for row in range(item0.rowCount()):
                    item1 = item0.child(row, column)
                    if item1 and item1.data(Qt.EditRole) == text:
                        found.append(item1)

        return found

    # ==================================================
    def set_check_state(self, item, row_data):
        """
        Set check state of row.

        Args:
            item (QItemModel): item of row(s).
            row_data (list): row data, [str].
        """
        for column in range(self.columnCount()):
            itemc = item[column]
            if self.column_type[column] == "check":
                itemc.setCheckable(True)
                if self.column_default[column] == "":
                    itemc.setEditable(False)
                state = str(row_data[column + 1]) == "True"
                if state:
                    itemc.setCheckState(Qt.Checked)
                else:
                    itemc.setCheckState(Qt.Unchecked)

    # ==================================================
    def get_row_data(self, index, column=None):
        """
        Get row data.

        Args:
            index (QModelIndex): index.
            column (int, optional): column.

        Returns:
            - (list) -- row data.

        Note:
            - bool string is replaced by bool.
        """
        if column is None:
            row_data = [self.itemFromIndex(index.siblingAtColumn(c)).data(Qt.EditRole) for c in range(self.columnCount())]
            row_data = [i == "True" if self.column_type[c] == "bool" else i for c, i in enumerate(row_data)]
        else:
            row_data = self.itemFromIndex(index.siblingAtColumn(column)).data(Qt.EditRole)
            if self.column_type[column] == "bool":
                row_data = row_data == "True"
        return row_data

    # ==================================================
    def emit_update_data(self, name, row_data, role, index):
        """
        Emit update data.

        Args:
            name (str): group name.
            row_data (list): row_data.
            role (int): role.
            index (QModelIndex): index.
        """
        if role in [Qt.EditRole, GroupModel.AppendRow]:
            self.dataModified.emit(name, row_data, index)
        elif role in [GroupModel.RemoveRow]:
            self.dataRemoved.emit(name, row_data, index)
        elif role in [Qt.CheckStateRole]:
            self.checkChanged.emit(name, row_data, index)

    # ==================================================
    def append_row(self, row_data=None, role=None, index=QModelIndex(), copy_row=False):
        """
        Append row.

        Args:
            row_data (list, optional): row data.
            role (int, optional): role.
            index (QModelIndex, optional): index.
            copy_row (bool, optional): copy from current index ?
        """
        if role is None:
            role = GroupModel.AppendRow

        # default.
        if row_data is None:
            row_data = copy.deepcopy(self.column_default)

        if len(row_data) != self.columnCount():
            return

        # insert copy or default.
        if index != QModelIndex():
            if copy_row:
                row_data = self.get_row_data(index)
                row_data[COLUMN_NAME_ACTOR] = ""
                row_data[COLUMN_LABEL_ACTOR] = ""
            else:
                row_data[0] = self.get_row_data(index, 0)

        name = row_data[0]  # assume tuple at first column.
        parent_item = self.find_item(name, child=False)
        if len(parent_item) > 0:
            parent_item = parent_item[0]
        else:
            parent_item = None

        if parent_item is None:  #  new group.
            item = [QStandardItem(str(i)) for i in row_data]
            self.set_check_state(item, row_data)
            self.appendRow(item)
            self.updateData.emit(self.group_name, row_data, role, item[0].index())
        else:  # existing group.
            if not parent_item.hasChildren():  # no children.
                row = parent_item.row()
                row_data0 = [self.item(row, i).data(Qt.EditRole) for i in range(self.columnCount())]
                item = [QStandardItem(str(i)) for i in row_data0]
                self.set_check_state(item, row_data0)
                parent_item.appendRow(item)
            item = [QStandardItem(str(i)) for i in row_data]
            self.set_check_state(item, row_data)
            parent_item.appendRow(item)
            self.updateData.emit(self.group_name, row_data, role, item[0].index())

        if self._update_widget:
            self.updateWidget.emit()

    # ==================================================
    def remove_row(self, index, role=None):
        """
        Remove row.

        Args:
            index (QModelIndex): index.
            role (int, optional): role.
        """
        if not index.isValid():
            return
        if role is None:
            role = GroupModel.RemoveRow

        index = index.siblingAtColumn(0)
        item = self.itemFromIndex(index)

        if item.parent():  # child.
            row = index.row()
            n = item.parent().rowCount()
            if row == 0:  # copy 2nd child to parent.
                row_data = self.get_row_data(item.parent().child(1).index())
                for c, d in enumerate(row_data):
                    super().setData(self.item(item.parent().row(), c).index(), d)
            self.updateData.emit(self.group_name, self.get_row_data(index), role, index)
            if n == 2:
                if row == 1:  # copy 1st child to parent.
                    row_data = self.get_row_data(item.parent().child(0).index())
                    for c, d in enumerate(row_data):
                        super().setData(self.item(item.parent().row(), c).index(), d)
                self.removeRow(0, index.parent())  # remove 2nd child.
                self.removeRow(0, index.parent())  # remove first child.
            else:
                self.removeRow(row, index.parent())  # remove row child.
        else:  # parent.
            n = self.rowCount(index)
            for row in range(n):
                cindex = self.index(row, 0, index)
                self.updateData.emit(self.group_name, self.get_row_data(cindex), role, cindex)
            if n == 0:
                self.updateData.emit(self.group_name, self.get_row_data(index), role, index)
            self.removeRow(index.row(), index.parent())

        if self._update_widget:
            self.updateWidget.emit()

    # ==================================================
    def move_row(self, index, value):
        """
        Move row.

        Args:
            index (QModelIndex): index.
            value (QVariant): value.
        """
        cols = range(self.columnCount())
        if index.parent() == QModelIndex():  # parent.
            item = self.itemFromIndex(index)
            if item.hasChildren():
                to_move = []
                for row in range(item.rowCount()):
                    row_data = [item.child(row, c).data(Qt.EditRole) for c in cols]
                    row_data[0] = value
                    to_move.append(row_data)
                self.remove_row(index, GroupModel.MoveRow)
                for row_data in to_move:
                    self.append_row(row_data=row_data, role=GroupModel.MoveRow)
            else:
                to_move = [self.invisibleRootItem().child(index.row(), c).data(Qt.EditRole) for c in cols]
                to_move[0] = value
                self.remove_row(index, GroupModel.MoveRow)
                self.append_row(row_data=to_move, role=GroupModel.MoveRow)
        else:  # child.
            item = self.itemFromIndex(index.parent())
            to_move = [item.child(index.row(), c).data(Qt.EditRole) for c in cols]
            to_move[0] = value
            self.remove_row(index, GroupModel.MoveRow)
            self.append_row(row_data=to_move, role=GroupModel.MoveRow)

    # ==================================================
    # action.
    # ==================================================
    def action_copy_row(self, indexes):
        """
        Slot for copy row action.

        Args:
            indexes (list): list of index.
        """
        if len(indexes) == 0:
            return

        indexes = [index for index in indexes if index.column() == 0]
        for index in indexes:
            self.append_row(index=index, copy_row=True)

    # ==================================================
    def action_insert_row(self, indexes):
        """
        Slot for insert row action.

        Args:
            indexes (list): list of index.

        Returns:
            - (QModelIndex) -- parent index of inserted row.
        """
        if len(indexes) == 0:
            self.append_row()
        else:
            indexes = [index for index in indexes if index.column() == 0]
            for index in indexes:
                self.append_row(index=index)

        # return parent index of inserted row.
        last = self.rowCount(QModelIndex()) - 1
        child = self.index(last, 0, QModelIndex())
        if self.rowCount(child) > 0:
            return child
        else:
            return QModelIndex()

    # ==================================================
    def action_remove_row(self, indexes):
        """
        Slot for remove row action.

        Args:
            indexes (list): list of index.
        """
        if len(indexes) == 0:
            return

        indexes = [index for index in indexes if index.column() == 0]

        for index in indexes:
            self.remove_row(index)

    # ==================================================
    # override.
    # ==================================================
    def setData(self, index, value, role=Qt.EditRole):
        """
        Set data (override).

        Args:
            index (QModelIndex): index.
            value (QVariant): value to set.
            role (int, optional): role.

        Returns:
            - (bool) -- success to set data ?
        """
        # in case of no change.
        if value == self.data(index, role):
            return False

        # move.
        if role == Qt.EditRole and index.column() == 0:
            # in order to close the editor, timer is used.
            QTimer.singleShot(0, lambda: self.move_row(index, value))
            return True

        # for child.
        item = self.itemFromIndex(index.siblingAtColumn(0))
        if item.hasChildren():
            for row in range(item.rowCount()):  # children.
                citem = item.child(row, index.column())
                if citem:
                    citem.setData(value, role)
                cindex = self.indexFromItem(citem)
                self.updateData.emit(self.group_name, self.get_row_data(cindex), role, cindex)
            status = super().setData(index, value, role)  # parent.
        else:
            status = super().setData(index, value, role)  # parent or child.
            self.updateData.emit(self.group_name, self.get_row_data(index), role, index)
            if not self.is_parent(index) and index.row() == 0:
                pindex = index.parent().siblingAtColumn(index.column())
                status = super().setData(pindex, value, role)

        return status

    # ==================================================
    # for debug.
    # ==================================================
    def debug_data_changed(self, topLeft, bottomRight, roles=[]):
        """
        Debug for dataChanged signal.

        Args:
            topLeft (QModelIndex): top left index.
            bottomRight (QModelIndex): bottom right index.
            roles (list, optional): list of roles.
        """
        roles = [self.get_role_str.get(i) for i in roles]

        print("roles:", roles)
        data = self.tolist_index(topLeft, bottomRight)
        for i in data:
            print("data:", i)

    # ==================================================
    def debug_updata_data(self, name, row_data, role):
        """
        Debug for updateData signal.

        Args:
            name (str): group name.
            row_data (list): row data.
            role (int): role.
        """
        role = self.get_role_str(role)
        print(role, name, row_data)

    # ==================================================
    def debug_item_tree(self):
        """
        Debug for showing item tree.
        """
        root_item = self.invisibleRootItem()

        s = ""
        for row in range(root_item.rowCount()):
            row_data = self.get_row_data(root_item.child(row).index())
            s += str(row_data) + "\n"
            p_item = root_item.child(row)
            for c_row in range(p_item.rowCount()):
                row_data = self.get_row_data(p_item.child(c_row).index())
                s += " " * 4 + str(row_data) + "\n"
        s = s[:-1]

        print(s)

    # ==================================================
    def show_index(self, index):
        """
        Get raw index.

        Args:
            index (QModelIndex): index.

        Returns:
            - (str) -- raw index, P/C(row,column).
        """
        p = "P" if self.is_parent(index) else "C"
        return f"{p}({index.row()},{index.column()})"

    # ==================================================
    def get_role_str(self, role):
        """
        Get role string.

        Args:
            role (int): role.

        Returns:
            - (str) -- role string.
        """
        # role string.
        role_str = {
            Qt.EditRole: "EditRole",
            Qt.DisplayRole: "DisplayRole",
            Qt.UserRole: "UserRole",
            Qt.CheckStateRole: "CheckStateRole",
            Qt.ToolTipRole: "ToolTipRole",
            Qt.StatusTipRole: "StatusTipRole",
            Qt.FontRole: "FontRole",
            Qt.TextAlignmentRole: "TextAlignmentRole",
            Qt.BackgroundRole: "BackgroundRole",
            Qt.ForegroundRole: "ForegroundRole",
            Qt.DecorationRole: "DecorationRole",
            Qt.SizeHintRole: "SizeHintRole",
            Qt.ItemDataRole: "ItemDataRole",
            GroupModel.AppendRow: "AppendRow",
            GroupModel.RemoveRow: "RemoveRow",
            GroupModel.MoveRow: "MoveRow",
        }
        return role_str.get(role, f"UndefinedRole({role})")

    # ==================================================
    def tolist_index(self, topLeft, bottomRight):
        """
        Row data for given index.

        Args:
            topLeft (QModelIndex): index at top left.
            bottomRight (QModelIndex): index at buttom right.

        Returns:
            - (list) -- row data.
        """
        parent = self.itemFromIndex(topLeft.parent())
        if parent is None:
            parent = self.invisibleRootItem()

        data = [self.get_row(parent.child(row).index()) for row in range(topLeft.row(), bottomRight.row() + 1)]

        return data

    # ==================================================
    def clear_data(self):
        """
        Clear data with keeping header and column info.
        """
        self.block_update_widget(True)
        for row in reversed(range(self.rowCount())):
            index = self.index(row, 0)
            for row1 in reversed(range(self.rowCount(index))):
                cindex = self.index(row1, 0, index)
                self.remove_row(cindex)
            self.remove_row(index)
        self.block_update_widget(False)
